# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
from unittest import mock

import pytest

from apache_beam.options import pipeline_options

from klio import __version__ as klio_lib_version
from klio_core import __version__ as klio_core_version

from klio_exec import __version__ as klio_exec_version


@pytest.fixture
def options():
    return pipeline_options.PipelineOptions()


def _gcp_options():
    return {
        "project": "test-project",
        "region": "us-central1",
        "temp_location": "gs://temp/location",
        "staging_location": "gs://staging/location",
        "dataflow_endpoint": "api.example.com",
        "service_account_email": "sudo@example.com",
        "no_auth": False,
        "template_location": "gs://template/location",
        "enable_streaming_engine": True,
        "dataflow_kms_key": "some_key",
        "flexrs_goal": "SPEED_OPTIMIZED",
        "labels": ["foo=bar", "baz=bla"],
    }


def _worker_options():
    return {
        "subnetwork": "projects/foo/regions/bar/network",
        "worker_machine_type": "n1-standard-2",
        "disk_size_gb": 32,
        "autoscaling_algorithm": "NONE",
        "num_workers": 2,
        "max_num_workers": 10,
        "worker_disk_type": (
            "compute.googleapis.com/projects/test-project/regions/"
            "us-central1/diskTypes/pd-ssd"
        ),
        "use_public_ips": False,
        "min_cpu_platform": "Intel Milkyway",
        "dataflow_worker_jar": "/foo/bar.ajr",
        "worker_harness_container_image": "gcr.io/example/image",
        "experiments": [],
        "streaming": True,
        "none_param": None,
    }


def _all_options():
    all_options = _gcp_options().copy()
    all_options.update(_worker_options())
    return all_options


@pytest.fixture
def pipeline_options_from_dict(all_options):
    return pipeline_options.PipelineOptions().from_dictionary(all_options)


def _config():
    mock_config = mock.Mock()
    mock_config.job_name = "my-job"
    mock_output = mock.Mock()
    mock_output.topic = "my-topic"
    mock_config.version = 2

    mock_input = mock.Mock()
    mock_input.subscription = "my-subscription"
    # Batch configs
    mock_input.name = "file"
    mock_input.location = "gs://sigint/dummy-file.txt"
    mock_input.to_io_kwargs.return_value = {
        "location": "gs://sigint/dummy-file.txt",
    }
    mock_output.name = "file"
    mock_output.location = "gs://sigint/"
    mock_output.to_io_kwargs.return_value = {
        "location": "gs://sigint/",
    }
    mock_job_config = mock.Mock()
    mock_job_config.events.inputs = [mock_input]
    mock_job_config.events.outputs = [mock_output]
    mock_job_config.data.inputs = [mock_input]
    mock_job_config.data.outputs = [mock_output]

    mock_pipeline_options = mock.Mock()

    for option, value in _gcp_options().items():
        setattr(mock_pipeline_options, option, value)

    for option, value in _worker_options().items():
        setattr(mock_pipeline_options, option, value)

    mock_pipeline_options.experiments = []
    mock_pipeline_options.streaming = True

    mock_pipeline_options.as_dict.return_value = _all_options()

    mock_config.job_config = mock_job_config
    mock_config.pipeline_options = mock_pipeline_options

    return mock_config


# The import in the few lines below triggers metrics to be setup (via
# helper decorators that are pulled in), which for some reason, triggers
# the setting up of stackdriver logger (which triggers an API auth call)
# on GH workflows. So we try to patch that out.
patcher = mock.patch("googleapiclient.discovery.build")
patcher.start()

from klio_exec.commands import run  # noqa E402


@pytest.fixture
def gcp_options():
    return _gcp_options()


@pytest.fixture
def worker_options():
    return _worker_options()


@pytest.fixture
def all_options():
    return _all_options()


@pytest.fixture
def config():
    return _config()


@pytest.mark.parametrize("run_callable", ["run", "run_basic"])
def test_get_run_callable(monkeypatch, mocker, run_callable):
    mock_run_module = mocker.Mock()
    expected_run_callable = mocker.Mock()

    if run_callable == "run":
        mock_run_module.return_value.run = expected_run_callable
        mock_run_module.return_value.run_basic = None
    else:
        mock_run_module.return_value.run = None
        mock_run_module.return_value.run_basic = expected_run_callable

    monkeypatch.setattr(run.imp, "load_source", mock_run_module)

    kpipe = run.KlioPipeline("my-job", mocker.Mock(), mocker.Mock())
    actual_callable = kpipe._get_run_callable()

    assert expected_run_callable == actual_callable


@pytest.mark.parametrize("exception", [None, ImportError, IOError])
def test_get_run_callable_raises(mocker, monkeypatch, caplog, exception):
    mock_run_module = mocker.Mock()

    if not exception:
        mock_run_module.return_value.run = None
        mock_run_module.return_value.run_basic = None
    else:
        mock_run_module.side_effect = exception

    monkeypatch.setattr(run.imp, "load_source", mock_run_module)

    kpipe = run.KlioPipeline("my-job", mocker.Mock(), mocker.Mock())

    with pytest.raises(SystemExit):
        kpipe._get_run_callable()

    assert 1 == len(caplog.records)


@pytest.mark.parametrize(
    "image,tag,expected_image",
    (
        ("gcr.io/foo/image", None, "gcr.io/foo/image"),
        ("gcr.io/foo/image:latest", None, "gcr.io/foo/image:latest"),
        ("gcr.io/foo/image", "my-tag", "gcr.io/foo/image:my-tag"),
        ("gcr.io/foo/image:latest", "my-tag", "gcr.io/foo/image:my-tag"),
    ),
)
def test_get_image_tag(image, tag, expected_image):
    actual_image = run.KlioPipeline._get_image_tag(image, tag)
    assert expected_image == actual_image


@pytest.mark.parametrize("direct_runner", (True, False))
def test_write_run_effective_config(mocker, direct_runner):
    if direct_runner:
        expected_path = "/usr/local/klio-job-run-effective.yaml"
    else:
        expected_path = "/usr/src/app/klio-job-run-effective.yaml"

    mock_config = mocker.Mock()
    mock_runtime_conf = mocker.Mock()
    mock_runtime_conf.direct_runner = direct_runner
    m_open = mocker.mock_open()
    mock_open = mocker.patch("klio_exec.commands.run.open", m_open)

    kpipe = run.KlioPipeline("test-job", mock_config, mock_runtime_conf)
    kpipe._write_run_effective_config()

    mock_open.assert_called_once_with(expected_path, "w")
    mock_config.write_to_file.assert_called_once_with(mock_open.return_value)


@pytest.mark.parametrize("streaming", (True, False))
@pytest.mark.parametrize(
    "exp,setup_file, requirements_file",
    [
        ("beam_fn_api", None, None),
        (None, "setup.py", None),
        (None, "setup.py", "requirements.txt"),
        (None, None, "requirements.txt"),
    ],
)
def test_verify_packaging(
    exp, setup_file, requirements_file, streaming, mocker
):
    mock_config = mocker.Mock()
    mock_config.pipeline_options = mocker.Mock()
    mock_config.pipeline_options.experiments = [exp]
    mock_config.pipeline_options.setup_file = setup_file
    mock_config.pipeline_options.requirements_file = requirements_file
    mock_config.pipeline_options.streaming = streaming
    mock_path_exists = mocker.patch.object(os.path, "exists")
    mock_path_exists.return_value = True

    mock_write_run_effective_config = mocker.patch.object(
        run.KlioPipeline, "_write_run_effective_config"
    )

    kpipe = run.KlioPipeline("test-job", mock_config, mocker.Mock())

    if not streaming and not any([requirements_file, setup_file, exp]):
        with pytest.raises(SystemExit):
            kpipe._verify_packaging()
    else:
        kpipe._verify_packaging()

    if setup_file and not exp:
        mock_write_run_effective_config.assert_called_once()
    else:
        mock_write_run_effective_config.assert_not_called()


def test_verify_packaging_with_both_packagaing_systems_raises(mocker):
    mock_config = mocker.Mock()
    mock_config.pipeline_options = mocker.Mock()
    mock_config.pipeline_options.experiments = ["beam_fn_api"]
    mock_config.pipeline_options.setup_file = "setup.py"

    kpipe = run.KlioPipeline("test-job", mock_config, mocker.Mock())

    with pytest.raises(SystemExit):
        kpipe._verify_packaging()


@pytest.mark.parametrize(
    "has_setup_file, setup_file_exists", ((True, False), (False, False),)
)
def test_verify_packaging_for_batch_raises(
    mocker, caplog, has_setup_file, setup_file_exists,
):
    mock_config = mocker.Mock()
    mock_config.pipeline_options = mocker.Mock()
    mock_config.pipeline_options.streaming = False
    mock_config.pipeline_options.requirements_file = None
    mock_config.pipeline_options.experiments = []
    if has_setup_file:
        mock_config.pipeline_options.setup_file = "setup.py"

    mock_path_exists = mocker.patch.object(os.path, "exists")
    mock_path_exists.return_value = setup_file_exists

    kpipe = run.KlioPipeline("test-job", mock_config, mocker.Mock())

    with pytest.raises(SystemExit):
        kpipe._verify_packaging()

        assert 1 == len(caplog.records)


def test_verify_packaging_for_batch_warns(mocker, caplog):
    mock_config = mocker.Mock()
    mock_config.pipeline_options = mocker.Mock()
    mock_config.pipeline_options.streaming = False
    mock_config.pipeline_options.experiments = ["beam_fn_api"]
    mock_config.pipeline_options.setup_file = None
    mock_config.pipeline_options.requirements_file = None

    kpipe = run.KlioPipeline("test-job", mock_config, mocker.Mock())

    kpipe._verify_packaging()

    assert 1 == len(caplog.records)
    assert "WARNING" == caplog.records[0].levelname


@pytest.mark.parametrize(
    "setup_file,reqs_file",
    (
        (None, None),
        ("setup.py", None),
        (None, "job-requirements.txt"),
        ("setup.py", "job-requirements.txt"),
    ),
)
def test_set_setup_options(
    all_options, config, setup_file, reqs_file, mocker, monkeypatch
):
    here = os.path.abspath(".")

    if reqs_file:
        all_options["requirements_file"] = reqs_file
    if setup_file:
        all_options["setup_file"] = setup_file

    options = pipeline_options.PipelineOptions().from_dictionary(all_options)
    actual_setup_options = options.view_as(pipeline_options.SetupOptions)

    kpipe = run.KlioPipeline("test-job", config, mocker.Mock())

    kpipe._set_setup_options(options)

    if setup_file:
        expected_setup_file = os.path.join(here, setup_file)
        assert expected_setup_file == actual_setup_options.setup_file
    else:
        assert actual_setup_options.setup_file is None
    if reqs_file:
        expected_reqs_file = os.path.join(here, reqs_file)
        assert expected_reqs_file == actual_setup_options.requirements_file
    else:
        assert actual_setup_options.requirements_file is None


def test_set_debug_options(config, mocker):
    test_options = mocker.Mock()
    kpipe = run.KlioPipeline("test-job", config, mocker.Mock())
    kpipe._set_debug_options(test_options)


@pytest.mark.parametrize("direct_runner", (True, False))
def test_set_standard_options(
    all_options, config, direct_runner, mocker, monkeypatch
):
    all_options["runner"] = "dataflow"
    options = pipeline_options.PipelineOptions().from_dictionary(all_options)
    actual_std_options = options.view_as(pipeline_options.StandardOptions)

    runtime_conf = mocker.Mock(direct_runner=direct_runner)
    kpipe = run.KlioPipeline("test-job", config, runtime_conf)
    kpipe._set_standard_options(options)

    assert actual_std_options.streaming is True

    expected_runner = "direct" if direct_runner else "dataflow"
    assert expected_runner == actual_std_options.runner


@pytest.mark.parametrize("fn_api_enabled", [True, False])
def test_set_worker_options(
    pipeline_options_from_dict,
    worker_options,
    config,
    fn_api_enabled,
    mocker,
    monkeypatch,
):
    expected_opts = [
        "subnetwork",
        "disk_size_gb",
        "autoscaling_algorithm",
        "num_workers",
        "max_num_workers",
        "use_public_ips",
        "min_cpu_platform",
        "dataflow_worker_jar",
    ]

    if fn_api_enabled:
        monkeypatch.setattr(
            config.pipeline_options, "experiments", ["beam_fn_api"]
        )

    kpipe = run.KlioPipeline("test-job", config, mocker.Mock(image_tag="foo"))
    kpipe._set_worker_options(pipeline_options_from_dict)

    actual_worker_options = pipeline_options_from_dict.view_as(
        pipeline_options.WorkerOptions
    )
    for opt in expected_opts:
        expected_value = worker_options[opt]
        # The True/False values in worker opts represent flags in the options.
        # These values are set to None when you set the PipelineOptions from
        # a dictionary. Since beam uses argparse to set values, None represents
        # False for these flags.
        # https://docs.python.org/2/howto/argparse.html
        # https://github.com/apache/beam/blob/master/sdks/python/apache_beam/options/pipeline_options.py#L723
        if not expected_value:
            expected_value = None
        # getattr should explode when not setting a default value
        assert expected_value == getattr(actual_worker_options, opt)

    assert (
        worker_options["worker_machine_type"]
        == actual_worker_options.machine_type
    )
    assert (
        worker_options["worker_disk_type"] == actual_worker_options.disk_type
    )

    if fn_api_enabled:
        assert (
            "gcr.io/example/image:foo"
            == actual_worker_options.worker_harness_container_image
        )
    else:
        assert (
            "gcr.io/example/image"
            == actual_worker_options.worker_harness_container_image
        )


@pytest.mark.parametrize(
    "update,exp_update,dataflow_endpoint,klio_cli_version,deployed_ci,user_env",
    [
        (True, True, None, None, False, True),
        (False, False, None, "1.2.3", True, True),
        (None, False, "api.example.com", None, False, False),
    ],
)
def test_set_google_cloud_options(
    all_options,
    config,
    update,
    exp_update,
    dataflow_endpoint,
    klio_cli_version,
    deployed_ci,
    user_env,
    gcp_options,
    mocker,
    monkeypatch,
):
    expected_opts = [
        "project",
        "region",
        "temp_location",
        "staging_location",
        "service_account_email",
        "no_auth",
        "template_location",
        "enable_streaming_engine",
        "dataflow_kms_key",
        "flexrs_goal",
    ]
    # this is to be changed when running `tox`; remove when no longer
    # supporting beam 2.14.0

    if dataflow_endpoint:
        all_options["dataflow_endpoint"] = dataflow_endpoint
    else:
        all_options.pop("dataflow_endpoint", None)

    options = pipeline_options.PipelineOptions().from_dictionary(all_options)

    actual_gcp_opts = options.view_as(pipeline_options.GoogleCloudOptions)

    monkeypatch.setattr(
        config.pipeline_options, "dataflow_endpoint", dataflow_endpoint
    )
    if klio_cli_version:
        monkeypatch.setenv("KLIO_CLI_VERSION", klio_cli_version)
        klio_cli_version_clean = klio_cli_version.replace(".", "-")
    if deployed_ci:
        monkeypatch.setenv("CI", "TRUE")
    if not user_env:
        monkeypatch.delenv("USER", raising=False)

    kpipe = run.KlioPipeline("test-job", config, mocker.Mock(update=update))
    kpipe._set_google_cloud_options(options)

    for opt in expected_opts:
        expected_value = gcp_options[opt]
        # getattr should explode when not setting a default value
        assert expected_value == getattr(actual_gcp_opts, opt)

    assert exp_update == actual_gcp_opts.update
    if dataflow_endpoint:
        assert dataflow_endpoint == actual_gcp_opts.dataflow_endpoint
    else:
        assert (
            "https://dataflow.googleapis.com"
            == actual_gcp_opts.dataflow_endpoint
        )
    user = None
    if deployed_ci:
        user = "CI"
    elif user_env:
        user = os.environ["USER"]

    klio_exec_version_clean = klio_exec_version.replace(".", "-")
    klio_core_version_clean = klio_core_version.replace(".", "-")
    klio_lib_version_clean = klio_lib_version.replace(".", "-")
    exp_labels = [
        "foo=bar",
        "baz=bla",
        "klio-exec={}".format(klio_exec_version_clean),
        "klio-core={}".format(klio_core_version_clean),
        "klio={}".format(klio_lib_version_clean),
    ]
    if user:
        exp_labels.append("deployed_by={}".format(user).lower())
    if klio_cli_version:
        exp_labels.append("klio-cli={}".format(klio_cli_version_clean))
    assert sorted(exp_labels) == sorted(actual_gcp_opts.labels)


def test_pipeline_options_from_dictionary(options):
    test_dict = {
        "random_param": "wombat",
        "project": "test_project",
        "subnetwork": "test_subnetwork",
        "num_workers": 2,
        "worker_machine_type": "test_machine_type",
        "use_public_ips": False,
    }

    test_options = options.from_dictionary(test_dict)
    google_options = test_options.view_as(pipeline_options.GoogleCloudOptions)
    worker_options = test_options.view_as(pipeline_options.WorkerOptions)

    assert "test_project" == google_options.project
    assert "test_subnetwork" == worker_options.subnetwork
    assert 2 == worker_options.num_workers
    assert "test_machine_type" == worker_options.machine_type
    assert worker_options.use_public_ips is None

    # This is testing that the options are created with the random param,
    # but the param is not accessible since it would not be recognized by
    # DataFlow.
    with pytest.raises(KeyError):
        test_options.get_all_options()["random_param"]


@pytest.mark.parametrize(
    "setup_file,reqs_file",
    ((None, None), ("setup.py", None), (None, "reqs.txt")),
)
def test_parse_config_pipeline_options(
    setup_file, reqs_file, all_options, config, mocker, monkeypatch
):
    as_dict_ret = config.pipeline_options.as_dict.return_value
    if setup_file:
        monkeypatch.setitem(as_dict_ret, "setup_file", setup_file)
        all_options["setup_file"] = setup_file
        all_options.pop("worker_harness_container_image")
    if reqs_file:
        monkeypatch.setitem(as_dict_ret, "requirements_file", reqs_file)
        all_options["requirements_file"] = reqs_file
        all_options.pop("worker_harness_container_image")
    kpipe = run.KlioPipeline("test-job", config, mocker.Mock())
    actual = kpipe._parse_config_pipeline_options()

    all_options.pop("none_param")

    assert all_options == actual


@pytest.mark.parametrize("has_none_values", [True, False])
def test_get_pipeline_options(
    has_none_values,
    config,
    options,
    worker_options,
    gcp_options,
    mocker,
    monkeypatch,
):

    mock_parse_config_pipeline_opts = mocker.Mock()
    mock_set_gcp_opts = mocker.Mock()
    mock_set_worker_opts = mocker.Mock()
    mock_set_std_opts = mocker.Mock()
    mock_set_debug_opts = mocker.Mock()
    mock_set_setup_opts = mocker.Mock()
    mock_opts = mocker.Mock()
    mock_opts_from_dict = mocker.Mock()

    monkeypatch.setattr(
        run.KlioPipeline,
        "_parse_config_pipeline_options",
        mock_parse_config_pipeline_opts,
    )

    monkeypatch.setattr(
        run.KlioPipeline, "_set_google_cloud_options", mock_set_gcp_opts
    )

    monkeypatch.setattr(
        run.KlioPipeline, "_set_worker_options", mock_set_worker_opts
    )

    monkeypatch.setattr(
        run.KlioPipeline, "_set_standard_options", mock_set_std_opts
    )

    monkeypatch.setattr(
        run.KlioPipeline, "_set_debug_options", mock_set_debug_opts
    )

    monkeypatch.setattr(
        run.KlioPipeline, "_set_setup_options", mock_set_setup_opts
    )

    monkeypatch.setattr(
        run.pipeline_options, "PipelineOptions", lambda: mock_opts
    )

    mock_opts_from_dict = mock_opts.from_dictionary.return_value

    mock_runtime_conf = mocker.Mock()

    kpipe = run.KlioPipeline("test-job", config, mock_runtime_conf)

    kpipe._get_pipeline_options()

    mock_parse_config_pipeline_opts.assert_called_once_with()

    mock_set_gcp_opts.assert_called_once_with(mock_opts_from_dict)
    mock_set_worker_opts.assert_called_once_with(mock_opts_from_dict)
    mock_set_std_opts.assert_called_once_with(mock_opts_from_dict)
    mock_set_debug_opts.assert_called_once_with(mock_opts_from_dict)
    mock_set_setup_opts.assert_called_once_with(mock_opts_from_dict)


@pytest.mark.parametrize(
    "direct_runner,run_error,exp_call_count",
    (
        (True, None, 1),
        (False, None, 1),
        (True, ValueError("No running job found with name"), 2),
    ),
)
@pytest.mark.parametrize("blocking", (True, False))
@pytest.mark.parametrize("streaming", (True, False))
def test_run_pipeline(
    streaming,
    blocking,
    direct_runner,
    run_error,
    exp_call_count,
    config,
    mocker,
    monkeypatch,
):
    job_name = "my-job"
    mock_runtime_config = mocker.Mock(
        direct_runner=direct_runner, update=True, blocking=blocking
    )

    mock_verify_packaging = mocker.Mock()
    mock_get_run_callable = mocker.Mock()
    mock_run_callable = mocker.Mock()
    mock_run_callable.return_value.__or__ = mocker.Mock()
    mock_get_run_callable.return_value = mock_run_callable
    mock_get_pipeline_options = mocker.Mock()
    mock_pipeline = mocker.Mock()
    mock_pipeline.return_value.__or__ = mocker.Mock()
    mock_read_from_pubsub = mocker.Mock()
    mock_read_from_pubsub.return_value.__or__ = mocker.Mock()
    mock_read_from_file = mocker.Mock()
    mock_write_to_file = mocker.Mock()
    mock_write_to_pubsub = mocker.Mock()
    mock_write_to_pubsub.return_value.__or__ = mocker.Mock()
    monkeypatch.setattr(
        run.BatchEventMapper, "input", {"file": mock_read_from_file},
    )
    monkeypatch.setattr(
        run.BatchEventMapper, "output", {"file": mock_write_to_file},
    )
    monkeypatch.setattr(
        run.StreamingEventMapper, "input", {"pubsub": mock_read_from_pubsub},
    )
    monkeypatch.setattr(
        run.StreamingEventMapper, "output", {"pubsub": mock_write_to_pubsub},
    )
    monkeypatch.setattr(
        run.KlioPipeline, "_verify_packaging", mock_verify_packaging
    )
    monkeypatch.setattr(
        run.KlioPipeline, "_get_run_callable", mock_get_run_callable
    )
    monkeypatch.setattr(
        run.KlioPipeline, "_get_pipeline_options", mock_get_pipeline_options
    )
    monkeypatch.setattr(run.beam, "Pipeline", mock_pipeline)
    monkeypatch.setattr(run.beam.io, "ReadFromPubSub", mock_read_from_pubsub)
    monkeypatch.setattr(run.beam.io, "WriteToPubSub", mock_write_to_pubsub)
    if streaming:
        mock_input = mocker.Mock()
        mock_input.name = "pubsub"
        mock_input.to_io_kwargs.return_value = {
            "subscription": "projects/foo/subscriptions/bar",
        }
        mock_output = mocker.Mock()
        mock_output.name = "pubsub"
        mock_output.to_io_kwargs.return_value = {
            "topic": "projects/foo/topics/bar",
        }
        config.job_config.events.inputs = [mock_input]
        config.job_config.events.outputs = [mock_output]

    if run_error:
        mock_pipeline.return_value.run.side_effect = [
            run_error,
            mock_pipeline.return_value.run.return_value,
        ]

    config.pipeline_options.streaming = streaming

    kpipe = run.KlioPipeline(job_name, config, mock_runtime_config)

    kpipe.run()

    assert exp_call_count == mock_verify_packaging.call_count
    mock_verify_packaging.assert_called_with()

    assert exp_call_count == mock_get_run_callable.call_count
    mock_get_run_callable.assert_called_with()

    assert exp_call_count == mock_get_pipeline_options.call_count
    mock_get_pipeline_options.assert_called_with()

    assert (
        exp_call_count
        == mock_get_pipeline_options.return_value.view_as.call_count
    )
    mock_get_pipeline_options.return_value.view_as.assert_called_with(
        pipeline_options.SetupOptions
    )

    assert exp_call_count == mock_pipeline.call_count
    mock_pipeline.assert_called_with(
        options=mock_get_pipeline_options.return_value
    )

    assert exp_call_count == mock_pipeline.return_value.run.call_count

    if direct_runner or blocking:
        result = mock_pipeline.return_value.run.return_value
        result.wait_until_finish.assert_called_once_with()


@pytest.mark.parametrize(
    "update,value_err_msg",
    (
        (False, "No running job found with name"),
        (True, "A different error"),
        (False, "A different error"),
    ),
)
def test_run_pipeline_raises(
    update, value_err_msg, config, mocker, monkeypatch, caplog
):
    job_name = "my-job"
    mock_runtime_config = mocker.Mock(update=update)

    mock_verify_packaging = mocker.Mock()
    mock_get_run_callable = mocker.Mock()
    mock_run_callable = mocker.Mock()
    mock_run_callable.return_value.__or__ = mocker.Mock()
    mock_get_run_callable.return_value = mock_run_callable
    mock_get_pipeline_options = mocker.Mock()
    mock_pipeline = mocker.Mock()
    mock_pipeline.return_value.__or__ = mocker.Mock()
    mock_read_from_pubsub = mocker.Mock()
    mock_read_from_pubsub.return_value.__or__ = mocker.Mock()
    mock_read_from_file = mocker.Mock()
    mock_read_from_file.return_value.__or__ = mocker.Mock()
    mock_write_to_file = mocker.Mock()
    mock_write_to_file.return_value.__or__ = mocker.Mock()
    mock_write_to_pubsub = mocker.Mock()
    mock_write_to_pubsub.return_value.__or__ = mocker.Mock()
    monkeypatch.setattr(
        run.KlioPipeline, "_verify_packaging", mock_verify_packaging
    )
    monkeypatch.setattr(
        run.KlioPipeline, "_get_run_callable", mock_get_run_callable
    )
    monkeypatch.setattr(
        run.KlioPipeline, "_get_pipeline_options", mock_get_pipeline_options
    )
    monkeypatch.setattr(run.beam, "Pipeline", mock_pipeline)
    monkeypatch.setattr(run.beam.io, "ReadFromPubSub", mock_read_from_pubsub)
    monkeypatch.setattr(
        run.transforms, "KlioReadFromText", mock_read_from_file
    )
    monkeypatch.setattr(run.transforms, "KlioWriteToText", mock_write_to_file)
    monkeypatch.setattr(run.beam.io, "WriteToPubSub", mock_write_to_pubsub)
    monkeypatch.setattr(
        run.BatchEventMapper, "input", {"file": mock_read_from_file},
    )
    monkeypatch.setattr(
        run.BatchEventMapper, "output", {"file": mock_write_to_file},
    )
    config.pipeline_options.streaming = False

    mock_pipeline.return_value.run.side_effect = ValueError(value_err_msg)

    kpipe = run.KlioPipeline(job_name, config, mock_runtime_config)
    with pytest.raises(SystemExit):
        kpipe.run()

    mock_pipeline.return_value.run.assert_called_once_with()
    assert 1 == len(caplog.records)
    assert "ERROR" == caplog.records[0].levelname



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import collections
import imp
import logging
import os
import re

import apache_beam as beam
from apache_beam.options import pipeline_options

from klio import __version__ as klio_lib_version
from klio import transforms
from klio.transforms import helpers
from klio_core import __version__ as klio_core_version
from klio_core import variables as var
from klio_core.config import core as config_core

from klio_exec import __version__ as klio_exec_version


DATAFLOW_LABEL_KEY_TO_OS_ENVIRON = {
    "build_id": "BUILD_ID",
    "organization": "ORGANIZATION",
    "repo": "REPOSITORY",
    "branch": "BRANCH_NAME",
    "commit_sha": "COMMIT_SHA",
    "klio-cli": "KLIO_CLI_VERSION",
}

RUN_CONFIG_PREAMBLE = """#
# THIS FILE IS AUTO-GENERATED BY KLIO
#
# To change your job's configuration, edit the job's klio-job.yaml or use the
# --config-file flag with klio-job-run to pass the path of a different
# configuration file to klio
#
"""

# Regex according to (https://cloud.google.com/resource-manager/
# docs/creating-managing-labels#requirements)
# otherwise, deployments will fail
DATAFLOW_LABEL_REGEX = re.compile(r"([a-zA-Z0-9_\-]+)")
HERE = os.path.abspath(".")


# NOTE: hopefully we don't get an dict lookup errors since KlioConfig
# should raise if given an unsupported event IO transform
class StreamingEventMapper(object):
    input = {"pubsub": transforms.KlioReadFromPubSub}
    output = {"pubsub": transforms.KlioWriteToPubSub}


class BatchEventMapper(object):
    input = {
        "file": transforms.KlioReadFromText,
        "bq": transforms.KlioReadFromBigQuery,
        "avro": transforms.KlioReadFromAvro,
    }
    output = {
        "file": transforms.KlioWriteToText,
        "bq": transforms.KlioWriteToBigQuery,
        "avro": transforms.KlioWriteToAvro,
    }


class EventIOMapper(object):
    streaming = StreamingEventMapper()
    batch = BatchEventMapper()


class KlioPipeline(object):
    def __init__(
        self, job_name, config, runtime_conf, event_io_mapper=EventIOMapper
    ):
        self.job_name = job_name
        self.config = config
        self.runtime_conf = runtime_conf
        if self.config.pipeline_options.streaming:
            self._io_mapper = event_io_mapper.streaming
        else:
            self._io_mapper = event_io_mapper.batch

    @property
    def _has_event_inputs(self):
        return len(self.config.job_config.events.inputs) > 0

    @property
    def _has_multi_event_inputs(self):
        return len(self.config.job_config.events.inputs) > 1

    @property
    def _has_event_outputs(self):
        return len(self.config.job_config.events.outputs) > 0

    @property
    def _has_data_inputs(self):
        return len(self.config.job_config.data.inputs) > 0

    @property
    def _has_multi_data_inputs(self):
        return len(self.config.job_config.data.inputs) > 1

    @property
    def _has_data_outputs(self):
        return len(self.config.job_config.data.outputs) > 0

    @property
    def _has_multi_data_outputs(self):
        return len(self.config.job_config.data.outputs) > 1

    def _set_setup_options(self, options):
        setup_options = options.view_as(pipeline_options.SetupOptions)

        if setup_options.setup_file:
            setup_file_path = os.path.join(HERE, setup_options.setup_file)
            setup_options.setup_file = setup_file_path

        if setup_options.requirements_file:
            reqs_file_path = os.path.join(
                HERE, setup_options.requirements_file
            )
            setup_options.requirements_file = reqs_file_path

    def _set_debug_options(self, options):
        pass

    def _set_standard_options(self, options):
        standard_opts = options.view_as(pipeline_options.StandardOptions)

        if self.runtime_conf.direct_runner:
            standard_opts.runner = "direct"

        elif (
            self.config.pipeline_options.runner
            == var.KlioRunner.DIRECT_GKE_RUNNER
        ):
            standard_opts.runner = (
                "klio_exec.runners.gke_direct.GkeDirectRunner"
            )

    @staticmethod
    def _get_image_tag(image, tag):
        if not tag:
            return image
        image_name = image.split(":")[0]
        return "{}:{}".format(image_name, tag)

    def _set_worker_options(self, options):

        worker_opts = options.view_as(pipeline_options.WorkerOptions)

        fnapi_enabled = (
            "beam_fn_api" in self.config.pipeline_options.experiments
        )
        if worker_opts.worker_harness_container_image and fnapi_enabled:
            image = KlioPipeline._get_image_tag(
                worker_opts.worker_harness_container_image,
                self.runtime_conf.image_tag,
            )
            worker_opts.worker_harness_container_image = image

    @staticmethod
    def _get_clean_label_value(label_value):
        # Get all regex matches of the label value, replace any unsuppoted
        # character with "-", and limit to 63 lowercase characters
        matches = re.findall(DATAFLOW_LABEL_REGEX, label_value)
        if not matches:
            return
        ret_label = "-".join(matches)
        ret_label = ret_label[:63]  # max 63 chars
        return ret_label.lower()

    def _set_google_cloud_options(self, options):
        gcp_opts = options.view_as(pipeline_options.GoogleCloudOptions)

        gcp_opts.job_name = self.job_name

        if self.runtime_conf.update is not None:
            gcp_opts.update = self.runtime_conf.update

        labels = gcp_opts.labels or []

        klio_exec_value = KlioPipeline._get_clean_label_value(
            klio_exec_version
        )
        klio_core_value = KlioPipeline._get_clean_label_value(
            klio_core_version
        )
        klio_value = KlioPipeline._get_clean_label_value(klio_lib_version)
        klio_labels = [
            "klio-exec={}".format(klio_exec_value),
            "klio-core={}".format(klio_core_value),
            "klio={}".format(klio_value),
        ]

        labels.extend(klio_labels)

        # Dataflow may not be able to handle duplicate keys; we should probably
        # do that here in some fashion (@lynn)
        for label, os_key in DATAFLOW_LABEL_KEY_TO_OS_ENVIRON.items():
            os_value = os.environ.get(os_key, "")
            os_value = KlioPipeline._get_clean_label_value(os_value)
            if os_value:
                labels.append("{}={}".format(label, os_value))

        deploy_user = os.environ.get("USER")
        if os.environ.get("CI", "").lower() == "true":
            # TODO: maybe provide way to allow something besides just "CI"
            deploy_user = "ci"

        if deploy_user:
            deploy_label = "deployed_by={}".format(
                KlioPipeline._get_clean_label_value(deploy_user)
            )

            labels.append(deploy_label)

        gcp_opts.labels = labels

    def _parse_config_pipeline_options(self):
        pipe_opts_dict = self.config.pipeline_options.as_dict()
        has_setup_file = pipe_opts_dict.get("setup_file") is not None
        has_reqs_file = pipe_opts_dict.get("requirements_file") is not None
        if any([has_setup_file, has_reqs_file]):
            # Dataflow will complain of not supporting custom images if
            # setup_file/reqs_file are used (w/o the beam_fn_api experiment)
            pipe_opts_dict.pop("worker_harness_container_image", None)
        return dict((k, v) for k, v in pipe_opts_dict.items() if v is not None)

    def _get_pipeline_options(self):
        # Remove None values since the from_dictionary sets these as
        # a string 'None' for the PipelineOptions flags.
        config_pipeline_options = self._parse_config_pipeline_options()

        options = pipeline_options.PipelineOptions().from_dictionary(
            config_pipeline_options
        )

        self._set_google_cloud_options(options)
        self._set_worker_options(options)
        self._set_standard_options(options)
        self._set_debug_options(options)
        self._set_setup_options(options)

        return options

    def _get_run_callable(self):
        run_path = "./run.py"
        try:
            run_module = imp.load_source("run", run_path)
            run_basic_callable = getattr(run_module, "run_basic", None)
            run_callable = getattr(run_module, "run", None)

            if not any([run_callable, run_basic_callable]):
                msg = "No 'run' function in run.py of job {}".format(
                    self.job_name
                )
                logging.error(msg)
                raise SystemExit(1)
            return run_basic_callable if run_basic_callable else run_callable
        except (ImportError, IOError):
            logging.error(
                "Could not import run.py in job {}".format(self.job_name),
                exc_info=True,
            )
            raise SystemExit(1)

    def _write_run_effective_config(self):
        # this method assumes setup.py is being used!
        if self.runtime_conf.direct_runner:
            path = config_core.WORKER_RUN_EFFECTIVE_CONFIG_PATH
        else:
            path = config_core.RUN_EFFECTIVE_CONFIG_PATH
        logging.debug(
            "Writing runtime configuration to {}"
            " in the job's running docker container.".format(path)
        )
        with open(path, "w") as f:
            f.write(RUN_CONFIG_PREAMBLE)
            self.config.write_to_file(f)

    def _verify_setup_py(self):
        # verify that setup.py has a reference to the runtime config file
        data_files_line = f'(".", ["{config_core.RUN_EFFECTIVE_CONFIG_FILE}"])'
        setup_file = self.config.pipeline_options.setup_file
        with open(setup_file, "r") as r:
            data = r.read()
            if config_core.RUN_EFFECTIVE_CONFIG_FILE not in data:
                logging.warning(
                    "Reference to 'klio-job-run-effective.yaml'"
                    " appears to be missing in 'setup.py'.  Please"
                    " ensure that the 'data_files' list includes "
                    f"the tuple: {data_files_line}."
                )

    def _verify_packaging(self):
        pipeline_opts = self.config.pipeline_options
        experiments = pipeline_opts.experiments
        fnapi_enabled = "beam_fn_api" in experiments
        has_setup_file = pipeline_opts.setup_file is not None
        setup_file_exists = has_setup_file and os.path.exists(
            pipeline_opts.setup_file
        )
        has_reqs_file = pipeline_opts.requirements_file is not None
        if fnapi_enabled and any([has_setup_file, has_reqs_file]):
            logging.error(
                "The 'beam_fn_api' experiment may not be enabled while "
                "providing a setup.py file and/or a requirements.txt file."
            )
            raise SystemExit(1)

        if not pipeline_opts.streaming:
            if fnapi_enabled:
                logging.warn(
                    "Support for batch jobs using the 'beam_fn_api' "
                    "experiment is still in development. "
                    "Use with caution."
                )

            if not any([fnapi_enabled, setup_file_exists]):
                if has_reqs_file:
                    logging.warn(
                        "Klio jobs are multi-module. "
                        "Thus, a setup.py file is required "
                        "in addition to a requirements.txt file"
                    )
                else:
                    logging.error(
                        "setup.py file either unspecified or not found."
                    )
                    raise SystemExit(1)

        if setup_file_exists:
            # when using setup.py, dump the current config to a runtime config
            # file to be included in the distribution package
            self._verify_setup_py()
            self._write_run_effective_config()

    def _setup_data_io_filters(self, in_pcol, label_prefix=None):
        # label prefixes are required for multiple inputs (to avoid label
        # name collisions in Beam)
        if self._has_multi_data_inputs or self._has_multi_data_outputs:
            logging.error(
                "Klio does not (yet) support multiple data inputs and outputs."
            )
            raise SystemExit(1)

        data_in_config, data_out_config = None, None
        if self._has_data_inputs:
            data_in_config = self.config.job_config.data.inputs[0]
        if self._has_data_outputs:
            data_out_config = self.config.job_config.data.outputs[0]

        pfx = ""
        if label_prefix is not None:
            pfx = "[{}] ".format(label_prefix)

        def lbl(label):
            return "{}{}".format(pfx, label)

        to_process_output = in_pcol
        pass_thru = None
        if data_in_config:
            pings = in_pcol | lbl("Ping Filter") >> helpers.KlioFilterPing()
            to_process_output = pings.process
            pass_thru = pings.pass_thru

        if data_out_config and not data_out_config.skip_klio_existence_check:
            output_exists = (
                to_process_output
                | lbl("Output Exists Filter")
                >> helpers.KlioGcsCheckOutputExists()
            )
            output_force = (
                output_exists.found
                | lbl("Output Force Filter") >> helpers.KlioFilterForce()
            )
            to_pass_thru_tuple = (pass_thru, output_force.pass_thru)
            to_pass_thru = (
                to_pass_thru_tuple
                | lbl("Flatten to Pass Thru") >> beam.Flatten()
            )

            to_filter_input_tuple = (
                output_exists.not_found,
                output_force.process,
            )
            to_filter_input = (
                to_filter_input_tuple
                | lbl("Flatten to Process") >> beam.Flatten()
            )
        else:
            to_pass_thru = pass_thru
            to_filter_input = to_process_output

        if data_in_config and not data_in_config.skip_klio_existence_check:
            input_exists = (
                to_filter_input
                | lbl("Input Exists Filter")
                >> helpers.KlioGcsCheckInputExists()
            )

            # TODO: update me to `var.KlioRunner.DIRECT_GKE_RUNNER` once
            #       direct_on_gke_runner_clean is merged
            if self.config.pipeline_options.runner == "DirectGKERunner":
                ack_inp_lbl = lbl("Ack Input Message from No Data Input Found")
                _ = input_exists.not_found | ack_inp_lbl >> beam.ParDo(
                    helpers.KlioAckInputMessage()
                )
            _ = (
                input_exists.not_found
                | lbl("Drop Not Found Data") >> helpers.KlioDrop()
            )
            to_process = input_exists.found
        else:
            to_process = to_filter_input

        return to_process, to_pass_thru

    def _update_audit_log(self, in_pcol, label_pfx=None):
        label = "Updating KlioMessage Audit Log"
        if label_pfx:
            label = "[{}] {}".format(label_pfx, label)

        return in_pcol | label >> helpers.KlioUpdateAuditLog()

    def _filter_intended_recipients(self, in_pcol, label_pfx=None):
        pfx = ""
        if label_pfx is not None:
            pfx = "[{}] ".format(label_pfx)

        def lbl(label):
            return "{}{}".format(pfx, label)

        # TODO: this "tagging by version then processing each version
        # differently" should only be temporary and removed once v2
        # migration is done
        version_lbl = lbl("Tag Message Versions")
        msg_version = in_pcol | version_lbl >> helpers._KlioTagMessageVersion()

        # tag each v1 message as 'process' or to 'drop' depending on if this
        # job should actually be handling the received message.
        v1_proc_lbl = lbl("Should Process v1 Message")
        v1_to_process = (
            msg_version.v1 | v1_proc_lbl >> helpers._KlioV1CheckRecipients()
        )
        v2_proc_lbl = lbl("Should Process v2 Message")
        v2_to_process = (
            msg_version.v2 | v2_proc_lbl >> helpers.KlioCheckRecipients()
        )

        flatten_ign_lbl = lbl("Flatten to Drop Messages to Ignore")
        to_drop_flatten = (v1_to_process.drop, v2_to_process.drop)
        to_drop = to_drop_flatten | flatten_ign_lbl >> beam.Flatten()

        # TODO: update me to `var.KlioRunner.DIRECT_GKE_RUNNER` once
        #       direct_on_gke_runner_clean is merged
        if self.config.pipeline_options.runner == "DirectGKERunner":
            ack_inp_lbl = lbl("Ack Dropped Input Message")
            _ = to_drop | ack_inp_lbl >> beam.ParDo(
                helpers.KlioAckInputMessage()
            )

        ignore_lbl = lbl("Drop Messages to Ignore")
        _ = to_drop | ignore_lbl >> helpers.KlioDrop()

        flatten_proc_lbl = lbl("Flatten to Process Intended Messages")
        to_process_flatten = (v1_to_process.process, v2_to_process.process)
        to_process = to_process_flatten | flatten_proc_lbl >> beam.Flatten()
        return to_process

    # TODO this can prob go away if/when we make event_inputs a
    # dictionary rather than a list of dicts (@lynn)
    def _generate_input_conf_names(self):
        ev_inputs = self.config.job_config.events.inputs
        input_dict = {}
        for index, ev in enumerate(ev_inputs):
            name = "{}{}".format(ev.name, index)
            input_dict[name] = ev
        return input_dict

    def _generate_pcoll_per_input(self, pipeline):
        inputs = self._generate_input_conf_names()
        MultiInputPCollTuple = collections.namedtuple(
            "MultiInputPCollTuple", list(inputs.keys())
        )
        input_name_to_input_pcolls = {}
        multi_to_pass_thru = []
        for input_name, input_conf in inputs.items():
            input_to_process, input_to_pass_thru = self._generate_pcoll(
                pipeline, input_conf, label_prefix=input_name
            )
            if input_to_pass_thru:
                multi_to_pass_thru.append(input_to_pass_thru)
            input_name_to_input_pcolls[input_name] = input_to_process

        to_process = MultiInputPCollTuple(**input_name_to_input_pcolls)
        to_pass_thru = (
            multi_to_pass_thru
            | "Merge multi-input pass-thrus" >> beam.Flatten()
        )
        return to_process, to_pass_thru

    def _generate_pcoll(self, pipeline, input_config, label_prefix=None):
        to_pass_thru = None
        to_process = pipeline

        if input_config.skip_klio_read:
            return to_process, to_pass_thru

        label = "Read Event Input"
        if label_prefix:
            label = "[{}] {}".format(label_prefix, label)

        transform_cls_in = self._io_mapper.input[input_config.name]
        in_pcol = pipeline | label >> transform_cls_in(
            **input_config.to_io_kwargs()
        )
        intended_msgs = self._filter_intended_recipients(in_pcol, label_prefix)
        audit_logged_msgs = self._update_audit_log(intended_msgs, label_prefix)
        to_process, to_pass_thru = self._setup_data_io_filters(
            audit_logged_msgs, label_prefix
        )
        return to_process, to_pass_thru

    # mutates the pipeline object, no need to return it
    def _setup_pipeline(self, pipeline):
        run_callable = self._get_run_callable()

        to_pass_thru = None
        to_process = pipeline
        # sanity check - I think klio config forces event input
        if self._has_event_inputs:
            if not self._has_multi_event_inputs:
                input_config = self.config.job_config.events.inputs[0]
                to_process, to_pass_thru = self._generate_pcoll(
                    pipeline, input_config
                )
            else:
                to_process, to_pass_thru = self._generate_pcoll_per_input(
                    pipeline
                )

        out_pcol = run_callable(to_process, self.config)

        # TODO: update me to `var.KlioRunner.DIRECT_GKE_RUNNER` once
        #       direct_on_gke_runner_clean is merged
        if self.config.pipeline_options.runner == "DirectGKERunner":
            if to_pass_thru:
                to_ack_input = (
                    out_pcol,
                    to_pass_thru,
                ) | "Flatten to Ack Input Messages" >> beam.Flatten()
            else:
                to_ack_input = out_pcol

            _ = to_ack_input | "Ack Input Messages" >> beam.ParDo(
                helpers.KlioAckInputMessage()
            )

        if self._has_event_outputs:
            output_config = self.config.job_config.events.outputs[0]
            if not output_config.skip_klio_write:
                transform_cls_out = self._io_mapper.output[output_config.name]
                to_output = out_pcol
                if to_pass_thru:
                    to_output_tuple = (out_pcol, to_pass_thru)
                    to_output = (
                        to_output_tuple | "Flatten to Output" >> beam.Flatten()
                    )

                _ = to_output | transform_cls_out(
                    **output_config.to_io_kwargs()
                )

    def run(self):
        self._verify_packaging()
        options = self._get_pipeline_options()
        options.view_as(pipeline_options.SetupOptions).save_main_session = True

        pipeline = beam.Pipeline(options=options)

        self._setup_pipeline(pipeline)

        try:
            # NOTE: When running with Dataflow, this `result` object has a lot
            #       of information about the job (id, name, project, status,
            #       etc). Could be useful if wanting to report back the status,
            #       URL of the dataflow job, etc.  @lynn
            result = pipeline.run()
        except ValueError as e:
            if (
                self.runtime_conf.update
                and "No running job found with name" in str(e)
            ):
                # job is currently not running - should simply deploy without
                # updating set
                # TODO: is this possible?
                self.runtime_conf = self.runtime_conf._replace(update=None)
                return self.run()

            logging.error("Error running pipeline: %s" % e)
            raise SystemExit(1)

        # TODO: update me to `var.KlioRunner.DIRECT_GKE_RUNNER` once
        #       direct_on_gke_runner_clean is merged
        is_direct_gke = (
            self.config.pipeline_options.runner == "DirectGKERunner"
        )
        should_block = (
            self.runtime_conf.direct_runner,
            self.runtime_conf.blocking,
            is_direct_gke,
        )
        if any(should_block):
            # the pipeline on direct runner will otherwise get garbage collected
            result.wait_until_finish()



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio_exec.commands.audit_steps import base


@pytest.fixture
def klio_config(mocker):
    conf = mocker.Mock()
    conf.pipeline_options = mocker.Mock()
    conf.pipeline_options.experiments = []
    return conf


@pytest.fixture
def mock_emit_warning(mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr(base.BaseKlioAuditStep, "emit_warning", mock)
    return mock


@pytest.fixture
def mock_emit_error(mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr(base.BaseKlioAuditStep, "emit_error", mock)
    return mock



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio_exec.commands.audit_steps import base


class DummyAuditStep(base.BaseKlioAuditStep):
    AUDIT_STEP_NAME = "dummy_step"

    @staticmethod
    def get_description():
        return "A description of a dummy step!"

    def after_tests(self):
        pass


class IncompleteAuditStep(base.BaseKlioAuditStep):
    AUDIT_STEP_NAME = "incomplete_dummy_step"


class NotAnAuditStep(object):
    def after_tests(self):
        pass


TB_PYTEST = [
    'File "/usr/local/lib/python3.6/site-packages/pluggy/callers.py", line 187, in _multicall\n  res = hook_impl.function(*args)',  # NOQA: E501
    'File "/usr/local/lib/python3.6/site-packages/_pytest/python.py", line 178, in pytest_pyfunc_call\n  testfunction(**testargs)',  # NOQA: E501
    'File "/usr/src/app/test_transform.py", line 13, in test_process\n  assert expected == list(output)[0]',  # NOQA: E501
    'File "/usr/src/app/transforms.py", line 21, in process\n  with tempfile.TemporaryFile() as t:',  # NOQA: E501
]
TB_NO_PYTEST = [
    'File "/usr/src/app/transforms.py", line 21, in process\n  with tempfile.TemporaryFile() as t:'  # NOQA: E501
]


@pytest.mark.parametrize("tb,exp_len", ((TB_PYTEST, 2), (TB_NO_PYTEST, 1)))
def test_remove_all_frames_until_after_pytest(tb, exp_len):
    act_ret = base._get_relevant_frames(tb)

    assert exp_len == len(act_ret)


def test_base_klio_audit_step(mock_terminal_writer):
    assert issubclass(DummyAuditStep, base.BaseKlioAuditStep)
    assert issubclass(IncompleteAuditStep, base.BaseKlioAuditStep)
    assert not issubclass(NotAnAuditStep, base.BaseKlioAuditStep)

    dummy_inst = DummyAuditStep("job/dir", "config", mock_terminal_writer)
    assert dummy_inst.before_tests() is None
    # just making sure this doesn't raise
    dummy_inst.after_tests()
    assert dummy_inst.get_description() is not None

    inc_inst = IncompleteAuditStep("job/dir", "config", mock_terminal_writer)
    assert inc_inst.before_tests() is None
    assert inc_inst.get_description() is None
    with pytest.raises(NotImplementedError):
        inc_inst.after_tests()


@pytest.mark.parametrize("tb", (None, ["a traceback"]))
def test_emit(tb, mock_terminal_writer, mocker, monkeypatch):
    mock_format_list = mocker.Mock()
    mock_format_list.return_value = tb
    monkeypatch.setattr(base.traceback, "format_list", mock_format_list)
    dummy_inst = DummyAuditStep("job/dir", "config", mock_terminal_writer)

    msg = "Emit this message"
    exp_msg = "[dummy_step]: Emit this message\n"
    if tb:
        exp_msg = "{}{}\n".format(exp_msg, tb[0])

    kwargs = {"foo": "bar"}

    assert dummy_inst.warned is False  # sanity check
    dummy_inst.emit_warning(msg, tb=tb, **kwargs)

    if tb:
        mock_format_list.assert_called_once_with(tb)
    else:
        mock_format_list.assert_not_called()
    exp_kwargs = {"foo": "bar", "yellow": True}
    mock_terminal_writer.write.assert_called_once_with(exp_msg, **exp_kwargs)
    assert dummy_inst.warned is True

    mock_format_list.reset_mock()
    mock_terminal_writer.reset_mock()

    assert dummy_inst.errored is False  # sanity check
    dummy_inst.emit_error(msg, tb=tb, **kwargs)

    if tb:
        mock_format_list.assert_called_once_with(tb)
    else:
        mock_format_list.assert_not_called()
    exp_kwargs = {"foo": "bar", "red": True}
    mock_terminal_writer.write.assert_called_once_with(exp_msg, **exp_kwargs)
    assert dummy_inst.errored is True



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio_exec.commands.audit_steps import multithreaded_tf


@pytest.mark.parametrize("tf_loaded", (True, False))
@pytest.mark.parametrize("worker_threads", (0, 1, 2))
def test_multithreaded_tf_usage(
    tf_loaded, worker_threads, klio_config, mock_emit_warning, mocker
):
    if worker_threads:
        klio_config.pipeline_options.experiments = [
            "worker_threads={}".format(worker_threads)
        ]

    if tf_loaded:
        mocker.patch.dict("sys.modules", {"tensorflow": ""})

    mt_tf_usage = multithreaded_tf.MultithreadedTFUsage(
        "job/dir", klio_config, "term_writer"
    )

    mt_tf_usage.after_tests()

    if worker_threads != 1 and tf_loaded:
        # don't care about the actual message
        assert 1 == mock_emit_warning.call_count
    else:
        mock_emit_warning.assert_not_called()



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import time

import line_profiler as lp
import psutil

from klio_exec.commands.utils import wrappers


class KLineProfiler(wrappers.KLineProfilerMixin, lp.LineProfiler):
    pass


# adapted from memory_profiler's approach:
# github.com/pythonprofilers/memory_profiler/blob/master/memory_profiler.py
def get_cpu_usage(proc, interval=None, stream=None):
    """Measure the cpu % usage at an interval."""
    if not interval:
        interval = 0.1
    if not stream:
        stream = sys.stdout

    line_count = 0
    while True:
        cpu_usage = psutil.cpu_percent()
        timestamp = time.time()
        # Same format as memory_profiler
        stream.write("CPU {:.1f} {:.4f}\n".format(cpu_usage, timestamp))

        time.sleep(interval)
        line_count += 1
        # flush every 50 lines. Make 'tail -f' usable on profile file
        if line_count > 50:
            line_count = 0
            stream.flush()

        if proc.poll() is not None:
            break



# Copyright 2021 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for GkeDirectRunner.

Adapted from apache_beam/runners/direct/direct_runner_test.py
"""
import logging
import threading

import apache_beam as beam
import hamcrest as hc
from apache_beam import pipeline
from apache_beam.metrics import cells
from apache_beam.metrics import execution
from apache_beam.metrics import metric
from apache_beam.metrics import metricbase
from apache_beam.testing import test_pipeline
from apache_beam.testing import util

from klio_exec.runners import gke_direct


def test_gke_direct_runs(caplog):
    runner = "klio_exec.runners.gke_direct.GkeDirectRunner"
    pipeline = test_pipeline.TestPipeline(runner=runner)
    _ = pipeline | beam.Create([{"foo": "bar"}])
    result = pipeline.run()
    result.wait_until_finish()

    gke_logger_records = [
        r
        for r in caplog.records
        if r.levelno == logging.INFO and r.name == "klio.gke_direct_runner"
    ]
    assert 1 == len(gke_logger_records)


def test_waiting_on_result_stops_executor_threads():
    pre_test_threads = set(t.ident for t in threading.enumerate())

    runner = "klio_exec.runners.gke_direct.GkeDirectRunner"
    pipeline = test_pipeline.TestPipeline(runner=runner)
    _ = pipeline | beam.Create([{"foo": "bar"}])
    result = pipeline.run()
    result.wait_until_finish()

    post_test_threads = set(t.ident for t in threading.enumerate())
    new_threads = post_test_threads - pre_test_threads
    assert 0 == len(new_threads)


def test_direct_runner_metrics():
    class MyDoFn(beam.DoFn):
        def start_bundle(self):
            count = metric.Metrics.counter(self.__class__, "bundles")
            count.inc()

        def finish_bundle(self):
            count = metric.Metrics.counter(self.__class__, "finished_bundles")
            count.inc()

        def process(self, element):
            gauge = metric.Metrics.gauge(self.__class__, "latest_element")
            gauge.set(element)
            count = metric.Metrics.counter(self.__class__, "elements")
            count.inc()
            distro = metric.Metrics.distribution(
                self.__class__, "element_dist"
            )
            distro.update(element)
            return [element]

    p = pipeline.Pipeline(gke_direct.GkeDirectRunner())
    pcoll = (
        p
        | beam.Create([1, 2, 3, 4, 5], reshuffle=False)
        | "Do" >> beam.ParDo(MyDoFn())
    )
    util.assert_that(pcoll, util.equal_to([1, 2, 3, 4, 5]))
    result = p.run()
    result.wait_until_finish()
    metrics = result.metrics().query()
    namespace = "{}.{}".format(MyDoFn.__module__, MyDoFn.__name__)

    hc.assert_that(
        metrics["counters"],
        hc.contains_inanyorder(
            execution.MetricResult(
                execution.MetricKey(
                    "Do", metricbase.MetricName(namespace, "elements")
                ),
                5,
                5,
            ),
            execution.MetricResult(
                execution.MetricKey(
                    "Do", metricbase.MetricName(namespace, "bundles")
                ),
                1,
                1,
            ),
            execution.MetricResult(
                execution.MetricKey(
                    "Do", metricbase.MetricName(namespace, "finished_bundles")
                ),
                1,
                1,
            ),
        ),
    )

    hc.assert_that(
        metrics["distributions"],
        hc.contains_inanyorder(
            execution.MetricResult(
                execution.MetricKey(
                    "Do", metricbase.MetricName(namespace, "element_dist")
                ),
                cells.DistributionResult(cells.DistributionData(15, 5, 1, 5)),
                cells.DistributionResult(cells.DistributionData(15, 5, 1, 5)),
            )
        ),
    )

    gauge_result = metrics["gauges"][0]
    hc.assert_that(
        gauge_result.key,
        hc.equal_to(
            execution.MetricKey(
                "Do", metricbase.MetricName(namespace, "latest_element")
            )
        ),
    )
    hc.assert_that(gauge_result.committed.value, hc.equal_to(5))
    hc.assert_that(gauge_result.attempted.value, hc.equal_to(5))



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio_exec.commands.utils import plugin_utils


@pytest.fixture
def mock_inspect(mocker, monkeypatch):
    mock = mocker.Mock()
    mock.getfile.return_value = "/path/to/package"
    monkeypatch.setattr(plugin_utils, "inspect", mock)
    return mock


@pytest.fixture
def mock_plugins(mocker):
    mock_dist = mocker.Mock()
    mock_dist.project_name = "package-name"
    mock_dist.parsed_version = "1.2.3"

    mock_plugins = []
    for i in range(2):
        mock_loaded = mocker.Mock()
        mock_loaded.AUDIT_STEP_NAME = "step-{}".format(i)

        mock_plugin = mocker.Mock(name=i)
        mock_plugin.load.return_value = mock_loaded
        mock_plugin.dist = mock_dist
        mock_plugins.append(mock_plugin)
    return mock_plugins


@pytest.fixture
def mock_iter_entry_points(mock_plugins, mocker, monkeypatch):
    mock = mocker.Mock()
    mock.return_value = mock_plugins
    monkeypatch.setattr(plugin_utils.pkg_resources, "iter_entry_points", mock)
    return mock


def test_load_plugins_by_namespace(mock_plugins, mock_iter_entry_points):
    ret = plugin_utils.load_plugins_by_namespace("a.namespace")

    mock_iter_entry_points.assert_called_once_with("a.namespace")
    for plugin in mock_plugins:
        plugin.load.assert_called_once_with()

    assert ret == [p.load.return_value for p in mock_plugins]


@pytest.mark.parametrize(
    "get_desc,docstr,exp_desc",
    (
        ("a description", None, "a description"),
        (None, "a docstring desc", "a docstring desc"),
        (None, None, "No description."),
        ("a description", "a docstring desc", "a description"),
    ),
)
def test_get_plugins_by_namespace(
    get_desc,
    docstr,
    exp_desc,
    mock_inspect,
    mock_plugins,
    mock_iter_entry_points,
    mocker,
):
    for mp in mock_plugins:
        mp.load.return_value.get_description.return_value = get_desc
        mp.load.return_value.__doc__ = docstr

    exp_plugins = []
    for i in range(2):
        exp_p = plugin_utils.KlioPlugin(
            plugin_name="step-{}".format(i),
            description=exp_desc,
            package_name="package-name",
            package_version="1.2.3",
            module_path="/path/to/package",
        )
        exp_plugins.append(exp_p)

    actual_plugins = list(
        plugin_utils._get_plugins_by_namespace("a.namespace")
    )

    mock_iter_entry_points.assert_called_once_with("a.namespace")
    for plugin in mock_plugins:
        plugin.load.assert_called_once_with()
        exp_calls = [mocker.call(plugin.load.return_value)]
        mock_inspect.getfile.assert_has_calls(exp_calls)

    assert len(mock_plugins) == mock_inspect.getfile.call_count
    assert exp_plugins == actual_plugins


@pytest.mark.parametrize("tw", (True, False))
def test_print_plugins(tw, mock_terminal_writer, mocker, monkeypatch):
    loaded_plugins = [
        plugin_utils.KlioPlugin(
            plugin_name="plugin_name",
            description="a desc",
            package_name="package-name",
            package_version="1.2.3",
            module_path="/path/to/package",
        )
    ]

    mock_get_plugins_by_ns = mocker.Mock()
    mock_get_plugins_by_ns.return_value = loaded_plugins
    monkeypatch.setattr(
        plugin_utils, "_get_plugins_by_namespace", mock_get_plugins_by_ns
    )

    exp_plugin_meta_str = " -- via package-name (v1.2.3) -- /path/to/package\n"
    exp_plugin_desc = "\ta desc\n\n"
    exp_calls = [
        mocker.call("plugin_name", blue=True, bold=True),
        mocker.call(exp_plugin_meta_str, green=True),
        mocker.call(exp_plugin_desc),
    ]

    kwargs = {"tw": None}
    if tw:
        kwargs["tw"] = mock_terminal_writer

    plugin_utils.print_plugins("a.namespace", **kwargs)

    mock_get_plugins_by_ns.assert_called_once_with("a.namespace")
    mock_terminal_writer.write.assert_has_calls(exp_calls)
    assert 3 == mock_terminal_writer.write.call_count



# Copyright 2021 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for KlioPubSubReadEvaluator

These were adapted from
apache_beam/io/gcp/pubsub_test.py::TestReadFromPubSub.

These validate that the original expected behavior from _PubSubReadEvaluator
was kept, as well checking that:
    * Responses are not auto-acked
    * MessageManager daemon threads started
    * Messages added to MessageManager
    * Messages handled one at a time, instead of 10 at a time
"""

import pytest

from apache_beam import transforms as beam_transforms
from apache_beam import utils as beam_utils
from apache_beam.io.gcp import pubsub as b_pubsub
from apache_beam.options import pipeline_options
from apache_beam.runners.direct import direct_runner
from apache_beam.runners.direct import transform_evaluator
from apache_beam.testing import test_pipeline as beam_test_pipeline
from apache_beam.testing import test_utils as beam_test_utils
from apache_beam.testing import util as beam_testing_util

from klio.message import pubsub_message_manager as pmm
from klio_core.proto import klio_pb2

from klio_exec.runners import evaluators


@pytest.fixture
def patch_msg_manager(mocker, monkeypatch):
    p = mocker.Mock(name="patch_msg_manager")
    monkeypatch.setattr(pmm, "MessageManager", p)
    return p


@pytest.fixture
def patch_sub_client(mocker, monkeypatch):
    # patch out network calls in SubscriberClient instantiation
    c = mocker.Mock(name="patch_sub_client")
    monkeypatch.setattr(evaluators.g_pubsub, "SubscriberClient", c)
    return c.return_value


class KlioTestPubSubReadEvaluator(object):
    """Wrapper of _PubSubReadEvaluator that makes it bounded."""

    _pubsub_read_evaluator = evaluators.KlioPubSubReadEvaluator

    def __init__(self, *args, **kwargs):
        self._evaluator = self._pubsub_read_evaluator(*args, **kwargs)

    def start_bundle(self):
        return self._evaluator.start_bundle()

    def process_element(self, element):
        return self._evaluator.process_element(element)

    def finish_bundle(self):
        result = self._evaluator.finish_bundle()
        result.unprocessed_bundles = []
        result.keyed_watermark_holds = {None: None}
        return result


transform_evaluator.TransformEvaluatorRegistry._test_evaluators_overrides = {
    direct_runner._DirectReadFromPubSub: KlioTestPubSubReadEvaluator,
}


def test_klio_pubsub_read_eval_read_messages_success(
    mocker, patch_sub_client, patch_msg_manager,
):

    exp_entity_id = "entity_id"
    kmsg = klio_pb2.KlioMessage()
    kmsg.data.element = bytes(exp_entity_id, "utf-8")
    data = kmsg.SerializeToString()

    publish_time_secs = 1520861821
    publish_time_nanos = 234567000
    attributes = {"key": "value"}
    ack_id = "ack_id"
    pull_response = beam_test_utils.create_pull_response(
        [
            beam_test_utils.PullResponseMessage(
                data, attributes, publish_time_secs, publish_time_nanos, ack_id
            )
        ]
    )
    pmsg = b_pubsub.PubsubMessage(data, attributes)
    expected_elements = [
        beam_testing_util.TestWindowedValue(
            pmsg,
            beam_utils.timestamp.Timestamp(1520861821.234567),
            [beam_transforms.window.GlobalWindow()],
        )
    ]
    patch_sub_client.pull.return_value = pull_response

    options = pipeline_options.PipelineOptions([])
    options.view_as(pipeline_options.StandardOptions).streaming = True
    with beam_test_pipeline.TestPipeline(options=options) as p:
        pcoll = p | b_pubsub.ReadFromPubSub(
            "projects/fakeprj/topics/a_topic", None, None, with_attributes=True
        )
        beam_testing_util.assert_that(
            pcoll,
            beam_testing_util.equal_to(expected_elements),
            reify_windows=True,
        )

    # Check overridden functionality:
    # 1. Check that auto-acking is skipped
    patch_sub_client.acknowledge.assert_not_called()
    # 2. Check that MessageManager daemon threads were started
    patch_msg_manager.assert_called_once_with(
        patch_sub_client.subscription_path()
    )
    # 3. Check that messages were added to the MessageManager
    patch_msg_manager.return_value.add.assert_called_once_with(ack_id, pmsg)
    # 4. Check that one message is handled at a time, instead of the
    #    original 10
    patch_sub_client.pull.assert_called_once_with(
        mocker.ANY, max_messages=1, return_immediately=True
    )

    patch_sub_client.api.transport.channel.close.assert_called_once_with()


def test_read_messages_timestamp_attribute_milli_success(
    mocker, patch_sub_client, patch_msg_manager,
):
    exp_entity_id = "entity_id"
    kmsg = klio_pb2.KlioMessage()
    kmsg.data.element = bytes(exp_entity_id, "utf-8")
    data = kmsg.SerializeToString()

    attributes = {"time": "1337"}
    publish_time_secs = 1520861821
    publish_time_nanos = 234567000
    ack_id = "ack_id"
    pull_response = beam_test_utils.create_pull_response(
        [
            beam_test_utils.PullResponseMessage(
                data, attributes, publish_time_secs, publish_time_nanos, ack_id
            )
        ]
    )
    pmsg = b_pubsub.PubsubMessage(data, attributes)
    expected_elements = [
        beam_testing_util.TestWindowedValue(
            pmsg,
            beam_utils.timestamp.Timestamp(
                micros=int(attributes["time"]) * 1000
            ),
            [beam_transforms.window.GlobalWindow()],
        ),
    ]
    patch_sub_client.pull.return_value = pull_response

    options = pipeline_options.PipelineOptions([])
    options.view_as(pipeline_options.StandardOptions).streaming = True
    with beam_test_pipeline.TestPipeline(options=options) as p:
        pcoll = p | b_pubsub.ReadFromPubSub(
            "projects/fakeprj/topics/a_topic",
            None,
            None,
            with_attributes=True,
            timestamp_attribute="time",
        )
        # Check original functionality that was kept the same
        beam_testing_util.assert_that(
            pcoll,
            beam_testing_util.equal_to(expected_elements),
            reify_windows=True,
        )

    # Check overridden functionality:
    # 1. Check that auto-acking is skipped
    patch_sub_client.acknowledge.assert_not_called()
    # 2. Check that MessageManager daemon threads were started
    patch_msg_manager.assert_called_once_with(
        patch_sub_client.subscription_path()
    )
    # 3. Check that messages were added to the MessageManager
    patch_msg_manager.return_value.add.assert_called_once_with(ack_id, pmsg)
    # 4. Check that one message is handled at a time, instead of the
    #    original 10
    patch_sub_client.pull.assert_called_once_with(
        mocker.ANY, max_messages=1, return_immediately=True
    )

    patch_sub_client.api.transport.channel.close.assert_called_once_with()


def test_read_messages_timestamp_attribute_rfc3339_success(
    mocker, patch_sub_client, patch_msg_manager,
):
    exp_entity_id = "entity_id"
    kmsg = klio_pb2.KlioMessage()
    kmsg.data.element = bytes(exp_entity_id, "utf-8")
    data = kmsg.SerializeToString()
    attributes = {"time": "2018-03-12T13:37:01.234567Z"}
    publish_time_secs = 1337000000
    publish_time_nanos = 133700000
    ack_id = "ack_id"
    pull_response = beam_test_utils.create_pull_response(
        [
            beam_test_utils.PullResponseMessage(
                data, attributes, publish_time_secs, publish_time_nanos, ack_id
            )
        ]
    )
    pmsg = b_pubsub.PubsubMessage(data, attributes)
    expected_elements = [
        beam_testing_util.TestWindowedValue(
            pmsg,
            beam_utils.timestamp.Timestamp.from_rfc3339(attributes["time"]),
            [beam_transforms.window.GlobalWindow()],
        ),
    ]
    patch_sub_client.pull.return_value = pull_response

    options = pipeline_options.PipelineOptions([])
    options.view_as(pipeline_options.StandardOptions).streaming = True
    with beam_test_pipeline.TestPipeline(options=options) as p:
        pcoll = p | b_pubsub.ReadFromPubSub(
            "projects/fakeprj/topics/a_topic",
            None,
            None,
            with_attributes=True,
            timestamp_attribute="time",
        )
        # Check original functionality that was kept the same
        beam_testing_util.assert_that(
            pcoll,
            beam_testing_util.equal_to(expected_elements),
            reify_windows=True,
        )

    # Check overridden functionality:
    # 1. Check that auto-acking is skipped
    patch_sub_client.acknowledge.assert_not_called()
    # 2. Check that MessageManager daemon threads were started
    patch_msg_manager.assert_called_once_with(
        patch_sub_client.subscription_path()
    )
    # 3. Check that messages were added to the MessageManager
    patch_msg_manager.return_value.add.assert_called_once_with(ack_id, pmsg)
    # 4. Check that one message is handled at a time, instead of the
    #    original 10
    patch_sub_client.pull.assert_called_once_with(
        mocker.ANY, max_messages=1, return_immediately=True
    )

    patch_sub_client.api.transport.channel.close.assert_called_once_with()


def test_read_messages_timestamp_attribute_missing(
    mocker, patch_sub_client, patch_msg_manager,
):
    exp_entity_id = "entity_id"
    kmsg = klio_pb2.KlioMessage()
    kmsg.data.element = bytes(exp_entity_id, "utf-8")
    data = kmsg.SerializeToString()

    attributes = {}
    publish_time_secs = 1520861821
    publish_time_nanos = 234567000
    publish_time = "2018-03-12T13:37:01.234567Z"
    ack_id = "ack_id"
    pull_response = beam_test_utils.create_pull_response(
        [
            beam_test_utils.PullResponseMessage(
                data, attributes, publish_time_secs, publish_time_nanos, ack_id
            )
        ]
    )
    pmsg = b_pubsub.PubsubMessage(data, attributes)
    expected_elements = [
        beam_testing_util.TestWindowedValue(
            pmsg,
            beam_utils.timestamp.Timestamp.from_rfc3339(publish_time),
            [beam_transforms.window.GlobalWindow()],
        ),
    ]
    patch_sub_client.pull.return_value = pull_response

    options = pipeline_options.PipelineOptions([])
    options.view_as(pipeline_options.StandardOptions).streaming = True
    with beam_test_pipeline.TestPipeline(options=options) as p:
        pcoll = p | b_pubsub.ReadFromPubSub(
            "projects/fakeprj/topics/a_topic",
            None,
            None,
            with_attributes=True,
            timestamp_attribute="nonexistent",
        )
        # Check original functionality that was kept the same
        beam_testing_util.assert_that(
            pcoll,
            beam_testing_util.equal_to(expected_elements),
            reify_windows=True,
        )

    # Check overridden functionality:
    # 1. Check that auto-acking is skipped
    patch_sub_client.acknowledge.assert_not_called()
    # 2. Check that MessageManager daemon threads were started
    patch_msg_manager.assert_called_once_with(
        patch_sub_client.subscription_path()
    )
    # 3. Check that messages were added to the MessageManager
    patch_msg_manager.return_value.add.assert_called_once_with(ack_id, pmsg)
    # 4. Check that one message is handled at a time, instead of the
    #    original 10
    patch_sub_client.pull.assert_called_once_with(
        mocker.ANY, max_messages=1, return_immediately=True
    )

    patch_sub_client.api.transport.channel.close.assert_called_once_with()


def test_read_messages_timestamp_attribute_fail_parse(patch_sub_client):
    exp_entity_id = "entity_id"
    kmsg = klio_pb2.KlioMessage()
    kmsg.data.element = bytes(exp_entity_id, "utf-8")
    data = kmsg.SerializeToString()

    attributes = {"time": "1337 unparseable"}
    publish_time_secs = 1520861821
    publish_time_nanos = 234567000
    ack_id = "ack_id"
    pull_response = beam_test_utils.create_pull_response(
        [
            beam_test_utils.PullResponseMessage(
                data, attributes, publish_time_secs, publish_time_nanos, ack_id
            )
        ]
    )
    patch_sub_client.pull.return_value = pull_response

    options = pipeline_options.PipelineOptions([])
    options.view_as(pipeline_options.StandardOptions).streaming = True
    p = beam_test_pipeline.TestPipeline(options=options)
    _ = p | b_pubsub.ReadFromPubSub(
        "projects/fakeprj/topics/a_topic",
        None,
        None,
        with_attributes=True,
        timestamp_attribute="time",
    )
    with pytest.raises(ValueError, match=r"parse"):
        p.run()

    patch_sub_client.acknowledge.assert_not_called()
    patch_sub_client.api.transport.channel.close.assert_called_with()



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tempfile

from klio_exec.commands.audit_steps import tempfile_usage


def test_tempfile_usage(klio_config, mock_emit_error, mocker, monkeypatch):
    inst = tempfile_usage.TempFileUsage("job/dir", klio_config, "tw")

    assert "TemporaryFile" == tempfile.TemporaryFile.__name__  # sanity check

    inst.before_tests()

    assert "MockTemporaryFile" == tempfile.TemporaryFile.__name__

    assert inst._tempfile_used is False
    inst.after_tests()
    mock_emit_error.assert_not_called()

    with tempfile.TemporaryFile():
        pass

    assert inst._tempfile_used is True

    inst.after_tests()
    # don't really care about message, just that it was called
    assert 1 == mock_emit_error.call_count



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio_exec.commands.utils import cpu_utils


@pytest.mark.parametrize("interval,exp_sleep", ((None, 0.1), (0.5, 0.5)))
def test_get_cpu_usage(interval, exp_sleep, mocker, monkeypatch, capsys):
    mock_proc = mocker.Mock()
    # wanting at least 50 lines to hit the flush
    proc_side_effect = [None] * 51
    proc_side_effect.append(True)
    mock_proc.poll.side_effect = proc_side_effect

    mock_time = mocker.Mock()
    timestamps = range(1, 53)
    mock_time.time.side_effect = timestamps
    monkeypatch.setattr(cpu_utils, "time", mock_time)

    cpu_measurements = (29.0, 51.4, 78.1, 61.2) * 13
    mock_cpu_percent = mocker.Mock(side_effect=cpu_measurements)
    monkeypatch.setattr(cpu_utils.psutil, "cpu_percent", mock_cpu_percent)

    cpu_utils.get_cpu_usage(mock_proc, interval=interval)

    assert 52 == mock_time.sleep.call_count
    assert mocker.call(exp_sleep) == mock_time.sleep.call_args

    captured = capsys.readouterr()

    data = zip(cpu_measurements, timestamps)
    exp_out = "\n".join("CPU {} {}.0000".format(d[0], d[1]) for d in data)
    exp_out = exp_out + "\n"
    assert exp_out == captured.out
